#!/usr/bin/env python3

import re
import xml.etree.ElementTree as etree  # noqa: N813
from typing import Any
from urllib.parse import unquote as urlunquote
from urllib.parse import urlparse

import markdown
from markdown.extensions import Extension
from markdown.inlinepatterns import InlineProcessor, SimpleTagInlineProcessor
from markdown.preprocessors import Preprocessor

from .config import Context

PRIO_BASE = 10000  # priority base for our extensions

# Patterns for ObsidianImages
OBSIMAGE_PATTERN = r'!\[\[(.*?)\]\]'
GENERICIMAGE_PATTERN = r'!\[(.*?)\]\((.*?)\)'

# Patterns for ObsidianLinks
OBSLINK_PATTERN = r'\[\[(.*?)\]\]'
GENERICLINK_PATTERN = r'\[(.*?)\]\((.*?)\)'


def is_proper_url(s: str) -> bool:
    """
    Checks to see if a string is a "proper" URL.

    Args:
        s (str): The string to be parsed.

    Returns:
        bool: ``True`` if the string is a proper URL, ``False`` if not.
    """
    parseout = urlparse(s)
    return True if parseout.scheme else False


def sanitize_reference(s: str) -> str:
    """
    Sanitizes an internal reference to a file by removing URL-quoted characters and any Markdown suffix.

    Args:
        s (str): Th reference to be sanitized.

    Returns:
        str: The sanitized reference.
    """
    rc = urlunquote(s)
    if rc.endswith('.md'):
        rc = rc[:-3]
    return rc


class MetaStripper(Extension):
    """
    An extension that strips the metadata off the front of Obsidian pages, as it's already been parsed in an
    earlier step.
    """

    class MetaStripperProc(Preprocessor):
        def run(self, lines: list[str]) -> list[str]:
            if lines[0] == '---':
                lines.pop(0)
                while lines[0] != '---':
                    lines.pop(0)
                lines.pop(0)
            return lines

    def extendMarkdown(self, md: markdown.Markdown) -> None:  # noqa: N802
        md.preprocessors.register(MetaStripper.MetaStripperProc(md), 'metastripper', PRIO_BASE)


class ObsidianImages(Extension):
    """An extension that supports image tags the way Obsidian handles them."""
    DIMS = re.compile(r'(.*)\|(\d+)(?:x(\d+))?')

    def __init__(self, context: Context, **kwargs: dict[str, Any]) -> None:
        super(ObsidianImages, self).__init__(**kwargs)
        self._context = context

    @property
    def invalid_reference_classname(self) -> str:
        return 'invalid-reference'

    def _parse_dimensions(self, s: str) -> tuple[str, int, int]:
        m = self.DIMS.match(s)
        if m:
            width = int(m.group(2))
            height = int(m.group(3)) if m.group(3) else -1
            return m.group(1), width, height
        else:
            return s, -1, -1

    def _lookup_image_reference(self, name: str) -> str | None:
        assert self._context.src_index is not None
        node, _ = self._context.src_index.lookup(name)
        if node:
            return node.link_target(self._context.url_prefix)
        return None

    class ObsidianImageProc(InlineProcessor):
        def __init__(self, pattern: str, md: markdown.Markdown, extref: Any) -> None:
            super(ObsidianImages.ObsidianImageProc, self).__init__(pattern, md)
            self._extref = extref

        def handleMatch(self, m: re.Match[str], data: str) -> tuple[etree.Element, int, int]:  # noqa: N802
            name, width, height = self._extref._parse_dimensions(m.group(1))
            link = self._extref._lookup_image_reference(name)
            if link is None:
                el = etree.Element('span')
                el.set('class', self._extref.invalid_reference_classname)
                el.text = name
            else:
                el = etree.Element('img')
                el.set('src', link)
                el.set('alt', name)
                if width > 0:
                    el.set('width', str(width))
                if height > 0:
                    el.set('height', str(height))
            return el, m.start(0), m.end(0)

    class GenericImageProc(InlineProcessor):
        def __init__(self, pattern: str, md: markdown.Markdown, extref: Any) -> None:
            super(ObsidianImages.GenericImageProc, self).__init__(pattern, md)
            self._extref = extref

        def handleMatch(self, m: re.Match[str], data: str) -> tuple[etree.Element, int, int]:  # noqa: N802
            name, width, height = self._extref._parse_dimensions(m.group(1))
            link = m.group(2)
            if is_proper_url(link):
                el = etree.Element('img')
                el.set('src', link)
                if len(name) > 0:
                    el.set('alt', name)
                if width > 0:
                    el.set('width', str(width))
                if height > 0:
                    el.set('height', str(height))
            else:
                newlink = self._extref._lookup_image_reference(link)
                if newlink:
                    el = etree.Element('img')
                    el.set('src', newlink)
                    if len(name) > 0:
                        el.set('alt', name)
                    if width > 0:
                        el.set('width', str(width))
                    if height > 0:
                        el.set('height', str(height))
                else:
                    el = etree.Element('span')
                    el.set('class', self._extref.invalid_reference_classname)
                    el.text = link
            return el, m.start(0), m.end(0)

    def extendMarkdown(self, md: markdown.Markdown) -> None:  # noqa: N802
        md.inlinePatterns.register(ObsidianImages.ObsidianImageProc(OBSIMAGE_PATTERN, md, self),
                                   'obsidian_images', PRIO_BASE + 1010)
        md.inlinePatterns.register(ObsidianImages.GenericImageProc(GENERICIMAGE_PATTERN, md, self),
                                   'obsidian_generic_images', PRIO_BASE + 1000)


class ObsidianLinks(Extension):
    """
    An extension that processes Obsidian internal links in the [[page name]] format, as well as overrides the standard
    Markdown link processing to handle Obsidian internal links as well as external links.
    """
    def __init__(self, context: Context, **kwargs: dict[str, Any]) -> None:
        super(ObsidianLinks, self).__init__(**kwargs)
        self._context = context

    @property
    def obsidian_link_classname(self) -> str:
        return 'obsidian-link'

    @property
    def invalid_reference_classname(self) -> str:
        return 'invalid-reference'

    def _parse_reference(self, contents: str) -> tuple[str | None, str]:
        text = None
        t = contents.split('|')
        if len(t) > 1:
            text = t[1]
            contents = t[0]

        assert self._context.src_index is not None
        node, _ = self._context.src_index.lookup(contents)
        if not text:
            text = contents
        if node:
            return node.link_target(self._context.url_prefix), text
        return None, text

    class ObsidianLinksProc(InlineProcessor):
        def __init__(self, pattern: str, md: markdown.Markdown, extref: Any) -> None:
            super(ObsidianLinks.ObsidianLinksProc, self).__init__(pattern, md)
            self._extref = extref

        def handleMatch(self, m: re.Match[str], data: str) -> tuple[etree.Element, int, int]:  # noqa: N802
            link, text = self._extref._parse_reference(m.group(1))
            if link is None:
                el = etree.Element('span')
                el.set('class', self._extref.invalid_reference_classname)
                el.text = text
            else:
                el = etree.Element('a')
                el.set('href', link)
                el.set('class', self._extref.obsidian_link_classname)
                el.text = text
            return el, m.start(0), m.end(0)

    class GenericLinksProc(InlineProcessor):
        def __init__(self, pattern: str, md: markdown.Markdown, extref: Any) -> None:
            super(ObsidianLinks.GenericLinksProc, self).__init__(pattern, md)
            self._extref = extref

        def handleMatch(self, m: re.Match[str], data: str) -> tuple[etree.Element, int, int]:  # noqa: N802
            text = m.group(1)
            link = m.group(2)
            if is_proper_url(link):
                el = etree.Element('a')
                el.set('href', link)
                el.text = text
            else:
                newlink, _ = self._extref._parse_reference(sanitize_reference(link))
                if newlink is None:
                    el = etree.Element('span')
                    el.set('class', self._extref.invalid_reference_classname)
                    el.text = text
                else:
                    el = etree.Element('a')
                    el.set('href', newlink)
                    el.set('class', self._extref.obsidian_link_classname)
                    el.text = text
            return el, m.start(0), m.end(0)

    def extendMarkdown(self, md: markdown.Markdown) -> None:  # noqa: N802
        md.inlinePatterns.register(ObsidianLinks.ObsidianLinksProc(OBSLINK_PATTERN, md, self),
                                   'obsidian_links', PRIO_BASE + 110)
        md.inlinePatterns.register(ObsidianLinks.GenericLinksProc(GENERICLINK_PATTERN, md, self),
                                   'obsidian_generic_links', PRIO_BASE + 100)


class ObsidianInlines(Extension):
    """An extension that handles the special Obsidian markdown format sequences."""
    def extendMarkdown(self, md: markdown.Markdown) -> None:  # noqa: N802
        md.inlinePatterns.register(SimpleTagInlineProcessor(r'()~~(.*?)~~', 'del'), 'strikeout', PRIO_BASE)
        md.inlinePatterns.register(SimpleTagInlineProcessor(r'()\=\=(.*?)\=\=', 'ins'), 'highlight', PRIO_BASE + 1)


def create_markdown_parser(context: Context) -> markdown.Markdown:
    return markdown.Markdown(extensions=[MetaStripper(),
                                         ObsidianImages(context),
                                         ObsidianLinks(context),
                                         ObsidianInlines()])
